first_experiment = """
Single Decision: The experiment consists of a single round in which you are matched with another randomly selected person. Note: You will only make a single set of decisions in this experiment.

Decisions: One of you will be designated as a Proposer (initial decision-maker) who will begin the round by proposing a division of an amount of money, $10.00. The other person, designated as responder, is told of the proposed division and must either accept it or not.

Role: You have been randomly assigned to be a Proposer (or initial decision-maker) in this process, and you will begin by suggesting amounts of money for each of you that sum to $10.00.

Earnings: If the responder accepts the proposed division, then each person earns their part of the proposed division. If the responder rejects, both earn $0.00.

Please remember that you are randomly matched with another participant, and that this process will not be repeated, you make only one decision.

The proposer (initial decision-maker) in each pair suggests a division of the $10.00; this proposal determines earnings for the round if it is accepted. A rejection produces earnings of $0.00 for each person. If the proposal is rejected, earnings are $0.00 for both proposer and responder.

After you and the person you are matched with have made your decisions, earnings will be determined and announced. There will not be another round of bargaining with this person or with anyone else.
"""

second_experiment = """
Single Decision: The experiment consists of a single round in which you are matched with another randomly selected person. Note: You will only make a single set of decisions in this experiment.

Initial Decisions: One of you will be designated as a Proposer (initial decision-maker) who will begin the round by proposing a division of an amount of money, $10.00. The other person, designated as responder, is told of the proposed division and must either accept it or not.

Second Stage Decisions: If the responder (second-mover) rejects, then the pie size is reduced to $3.00, and the responder can make a proposal for how to divide this lower amount. If a counter proposal is made, the initial proposer (first mover), must either accept it or not (no additional proposals would be allowed in this round).

Role: You have been randomly assigned to be a Proposer (or initial decision-maker) in this process, and you will begin by suggesting amounts of money for each of you that sum to $10.00.

Earnings: If the responder accepts the proposed division, then each person earns their part of the proposed division. If the responder rejects, then the amount of money to be divided becomes $3.00, and the responder must make a counter-proposal on how to divide this new amount. If the initial proposer accepts this counter-proposal, then it is implemented. If the counter-proposal is rejected, then each person earns $0.00.

Fixed Payments: In addition to the amounts earned (if any) from the bargaining process, each person may receive a fixed payment that is independent of the outcome of the bargaining outcome. Your fixed payment will be: $4.00 and the fixed payment of the other person will be $0.00. Note: These fixed payments are received each round, irrespective of the bargaining outcomes.

Please remember that you are randomly matched with another participant, and that this process will not be repeated, you make only one decision.

The proposer (initial decision-maker) in each pair suggests a division of the $10.00; this proposal determines earnings for the round if it is accepted.

If the initial proposal is rejected, the amount of money to be divided is changed to $3.00, and the responder must now make a counter-proposal for how to divide this new amount.

Then the initial proposer (first mover) must either accept the counter-proposal or reject. An acceptance implements the counter-proposal, and a rejection results in earnings of $0.00 (for both) from the bargaining process.

After you and the person you are matched with have made your decisions, earnings will be determined and announced. There will not be another round of bargaining with this person or with anyone else.

In addition to these earnings (if any), the initial proposer receives $4.00 and the responder receives $0.00 each round, indpendently of the other outcomes.

Bargaining Process: This process has changed.
Responder Decision (Accept or Make Counter-Proposal): If the responder accepts the initial proposal, it is implemented, but if the responder rejects, then the responder must make a counter-proposal to divide an amount of money, $3.00. Then the proposer must either accept or reject this counter-proposal. Final acceptance by the proposer implements the counter-proposal, and rejection results in zero earnings for both from the bargaining process.
Pie Size: The amount of money to be divided is unchanged at $10.00.
Fixed Payments: In addition to these earnings (if any), the proposer receives $4.00 and the responder receives $0.00 each round.
Matching Method: The proposer/responder pairs will remain the same for all rounds in this part.

In the first stage, you enter a proposed money amount for yourself, between $0.00 and $10.00. The remainder (between $0.00 and $10.00) is what is proposed for the responder. If your proposal is rejected, the amount of money available will be reduced to $3.00, and the other person will suggest amounts for each of you that sum to $3.00.
"""


import sys
from . import experiment as exp # multi round ultimatum game
from . import ultimatum_exp as ult # multi stage ultimatum game
import chromedriver_autoinstaller
import subprocess

# Define repository details
repo_dir = "./logs"

def upload():
    # Read the user's name and email from secrets.txt
    with open("secrets.txt", "r") as f:
        _ = f.readline().strip()
        username, token = f.readline().strip().split(" ")
    # Define the remote repository URL
    remote_url = f"https://{username}:{token}@github.com/JawandS/econ-llm-data.git"
    # Initialize the repository
    subprocess.run(["git", "init"], cwd=repo_dir, check=True)
    # Add the remote url
    subprocess.run(["git", "remote", "add", "origin", remote_url], cwd=repo_dir, check=True)
    # Add all files to the repository
    subprocess.run(["git", "add", "."], cwd=repo_dir, check=True)
    # Commit the changes
    subprocess.run(["git", "commit", "-m", "Upload results"], cwd=repo_dir, check=True)
    # Push the changes to the remote repository
    subprocess.run(["git", "push", "-u", "origin", "master"], cwd=repo_dir, check=True)
    # Print a success message
    print("Results uploaded successfully")

def main():
    # Automatically install the latest version of chromedriver
    chromedriver_autoinstaller.install()

    # Check if the user provided the correct number of arguments
    if len(sys.argv) < 2:
        print("Usage: econ-llm run [experiment_name] [user_id]")
        sys.exit(1)
    
    command = sys.argv[1]
    if command == "run":
        args = sys.argv[2:]
        print(f"Running with arguments: {args}")
        exp.run_experiment(*args)
    elif command == "ult": # ult stands for ultimatum
        print("Running ultimatum game experiment")
        ult.run_ultimatum_experiment()
    elif command == "upload": # upload the results to the server
        print("Uploading results to the server")
    else:
        print(f"Unknown command: {command}")
        sys.exit(1)

    print("Results upload started")
    upload()

if __name__ == "__main__":
    upload()


# Create a script that will automatically interact with veconlab

steps = """
- https://veconlab.econ.virginia.edu/login1.php
    - Enter: xaty4
    - Click submit
- https://veconlab.econ.virginia.edu/login2.php
    - First Name: [First Name]
    - Last Name: [Last Name]
    - Optional password: [Password]
    - Click continue to proceed
- https://veconlab.econ.virginia.edu/inst_0.php
    - Save the instructions as a variable
    - Click continue with instructions
- https://veconlab.econ.virginia.edu/bg/bg_inst1.php
    - Save the instructions as a variable
    - Click continue with instructions
- https://veconlab.econ.virginia.edu/bg/bg_inst2.php
    - Save the instructions as a variable
    - Click continue
- https://veconlab.econ.virginia.edu/bg/bg_inst6.php
    - Save the instructions as a variable
    - Click Finish with instructions
"""

from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import Select

from .agent import Agent
import json
import time
import os
import threading

class VeconLabAutomation:
    """
    A class to automate interactions with VeconLab using Selenium.
    """

    def __init__(self, session_id: str, first_name: str, last_name: str, password: str = ""):
        """
        Initialize the Selenium WebDriver and user credentials.

        :param session_id: The session ID for the game.
        :param first_name: User's first name.
        :param last_name: User's last name.
        :param password: Optional password for login.
        """
        self.session_id = session_id
        self.first_name = first_name
        self.last_name = last_name
        self.password = password
        self.instructions = []

        # Set up the WebDriver
        self.driver = webdriver.Chrome()
        self.wait = WebDriverWait(self.driver, timeout = 10 * 60) # wait ten minutes

        # Set up AI Agent
        context_text = """
        You are an undergraduate student at the college of William & Mary participating in an economics experiment for money. 
        Your role is to play an ultimatum game with another participant. You will have to make a series of decisions and your payoff is based on your opponent as well
        If the responder accepts the offer, both players receive the amount proposed by the proposer. If the responder rejects the offer, both players receive nothing.
        For all questions you should provide your answer in the following format: {"choice": "YOUR_CHOICE", "explanation": "YOUR_EXPLANATION"}.
        YOUR_CHOICE should be the minimum information needed to answer the question. For example, if asked for a number only put the number in YOUR_CHOICE. For a decision put either 'a' for accept or 'r' for reject.
        YOUR_EXPLANATION should be a short (1-2 sentence) justification for your choice.
        """
        self.agent = Agent(secrets_file="secrets.txt")
        self.agent.add_context(context_text)

        # Log
        print("VeconLabAutomation initialized.")

    def login_step_1(self):
        """
        Navigate to the first login page and submit the session ID.
        """
        url = "https://veconlab.econ.virginia.edu/login1.php"
        self.driver.get(url)

        input_field = self.driver.find_element(By.NAME, "table_name")
        submit_button = self.driver.find_element(By.XPATH, "//input[@type='submit']")
        
        input_field.send_keys(self.session_id)

        # Log
        print("Submitting session ID...")
        submit_button.click()

    def login_step_2(self):
        """
        Navigate to the second login page and enter user details.
        """
        self.wait.until(EC.url_contains("login2.php"))
        
        self.driver.find_element(By.NAME, "first_name").send_keys(self.first_name)
        self.driver.find_element(By.NAME, "last_name").send_keys(self.last_name)
        self.driver.find_element(By.NAME, "password").send_keys(self.password)
        self.driver.find_element(By.NAME, "password_check").send_keys(self.password)
        submit_button = self.driver.find_element(By.XPATH, "//input[@type='submit']")
        # Log
        print("Logging in...")
        submit_button.click()

    def save_instructions(self):
        """
        Save instructions from the current page and proceed.
        """
        font_elements = self.driver.find_elements(By.TAG_NAME, "font")
        for element in font_elements:
            instruction_text = element.text
            self.instructions.append(instruction_text)
        print("Moving past instructions...")
        submit_button = WebDriverWait(self.driver, 10).until(
            EC.presence_of_element_located((By.XPATH, "//input[@type='submit']"))
        )
        submit_button.click()

    def play_game(self):
        """
        Play the game and save the results.
        """
        def _play_game(id):
            # Get the prompt
            prompt_element = self.driver.find_element(By.TAG_NAME, "font")
            # Get the decision (select element)
            decision_element = self.driver.find_element(By.TAG_NAME, "select")
            # Print choices
            # options = [option.get_attribute("value") for option in decision_element.find_elements(By.TAG_NAME, "option")]
            # Select a random choice
            if id == 0:
                offer_json = json.loads(self.agent.prompt("You are the proposer. How much money do you keep for youself (between $1-10?). Your opponent will receive the rest."))
                choice = offer_json['choice']
                print(f"Selected: {choice} b/c {offer_json['explanation']}")
            elif id == 1:
                response_json = json.loads(self.agent.prompt(f"Your previous offer was rejected. You can make a new offer. How much money do you keep for youself (between $1-10?). Your opponent will receive the rest."))
                choice = response_json['choice'].strip()
                print(f"Selected: {choice} b/c {response_json['explanation']}")
            # Submit the choice
            select = Select(decision_element)
            select.select_by_value(choice)
            # Click submit
            print("Submitting decision...")
            self.driver.find_element(By.XPATH, "//input[@type='submit']").click()
            
            # Wait until new page loads
            self.wait.until(EC.url_contains("bg_p_confirm.php"))
            # Confirm decision
            self.wait.until(EC.presence_of_element_located((By.XPATH, "//input[@value='Confirm Decision']")))
            print("Confirming decision...")
            self.driver.find_element(By.XPATH, "//input[@value='Confirm Decision']").click()

        _play_game(id=0)

        # Wait until the results page loads
        self.wait.until(EC.url_contains("bg_p_results.php"))
        # Wait until a submit button appears
        self.wait.until(EC.presence_of_element_located((By.XPATH, "//input[@value='Begin Round 2']")))
        # Click submit
        print("Starting round 2...")
        self.driver.find_element(By.XPATH, "//input[@value='Begin Round 2']").click()

        # Clear instructions
        self.instructions = []
        # Go through instruction pages
        self.wait.until(EC.url_contains("bg_inst1.php"))
        self.save_instructions()
        self.wait.until(EC.url_contains("bg_inst2.php"))
        self.save_instructions()
        self.wait.until(EC.url_contains("bg_inst6.php"))
        self.save_instructions()

        # Play the game again
        _play_game(id=1)

        # Wait until the results page loads
        self.wait.until(EC.url_contains("bg_p_results.php"))
        # Wait until a submit button appears
        print("Waiting for counter offer...")
        self.wait.until(EC.presence_of_element_located((By.XPATH, "//input[@value='Respond to Counter-Proposal']")))
        # Click submit
        print("Responding to counter offer...")
        self.driver.find_element(By.XPATH, "//input[@value='Respond to Counter-Proposal']").click()

        # Select accept or reject
        print("Waiting on results...")
        self.wait.until(EC.url_contains("bg_p_submit.php"))
        # get counter offer from page
        counter_offer = 3 - int(self.driver.find_element(By.NAME, "r_counter_proposal").get_attribute("value"))
        print(f"counter offer: {counter_offer}")
        # prompt agent
        offer_json = json.loads(self.agent.prompt(f"Do you accept or reject the counter offer of ${counter_offer} for you?"))
        choice = offer_json['choice']
        print(f"Selected: {choice} b/c {offer_json['explanation']}")
        # Select radio button by value
        self.driver.find_element(By.XPATH, f"//input[@value='{choice}']").click()
        # Click submit
        print("Submitting decision...")
        self.driver.find_element(By.XPATH, "//input[@type='submit']").click()

        # Wait until the confirmation page loads
        self.wait.until(EC.url_contains("bg_p_confirm.php"))
        # Click
        self.wait.until(EC.presence_of_element_located((By.XPATH, "//input[@value='Confirm Decision']")))
        print("Confirming decision...")
        self.driver.find_element(By.XPATH, "//input[@value='Confirm Decision']").click()

    def respond_to_offer(self):
        """
        Automation pipelines for the responder.
        """
        # Wait until submit page
        print("Waiting for offer...")
        # Print HTML
        while "p_decision1" not in self.driver.page_source:
            # sleep 1 second
            time.sleep(1)
            print("Waiting on offer...")
        # self.wait.until(EC.presence_of_element_located((By.NAME, "//input[@value='p_decision1']")))
        # Get the offer
        offer = 10 - int(self.driver.find_element(By.NAME, "p_decision1").get_attribute("value"))
        print(f"Responder got offer of: {offer}")
        # Prompt agent
        offer_json = json.loads(self.agent.prompt(f"Do you accept or reject the offer of ${offer} for you?"))
        choice = offer_json['choice']
        print(f"Selected: {choice} b/c {offer_json['explanation']}")
        # Select radio button by value
        self.driver.find_element(By.XPATH, f"//input[@value='{choice}']").click()
        # Submit decision
        print("Submitting decision...")
        self.driver.find_element(By.XPATH, "//input[@value='Submit']").click()
        
        # Wait until confirmation page
        self.wait.until(EC.url_contains("bg_r_confirm.php"))
        # Confirm decision
        self.wait.until(EC.presence_of_element_located((By.XPATH, "//input[@value='Confirm Decision']")))
        print("Confirming decision...")
        self.driver.find_element(By.XPATH, "//input[@value='Confirm Decision']").click()

        # Load results page
        self.wait.until(EC.url_contains("bg_r_results.php"))
        # Begin round 2
        self.wait.until(EC.presence_of_element_located((By.XPATH, "//input[@value='Begin Round 2']")))
        print("Starting round 2...")
        self.driver.find_element(By.XPATH, "//input[@value='Begin Round 2']").click()

        # Clear instructions
        self.instructions = []
        # Go through instruction pages
        self.wait.until(EC.url_contains("bg_inst1.php"))
        self.save_instructions()
        self.wait.until(EC.url_contains("bg_inst2.php"))
        self.save_instructions()
        self.wait.until(EC.url_contains("bg_inst6.php"))
        self.save_instructions()


        # Wait until decision page
        # self.wait.until(EC.url_contains("bg_r_submit.php"))
        # Get new offer
        self.wait.until(EC.presence_of_element_located((By.NAME, "p_decision1")))
        offer = 10 - int(self.driver.find_element(By.NAME, "p_decision1").get_attribute("value"))
        print(f"New offer for responder is: {offer}")
        # Prompt agent
        offer_json = json.loads(self.agent.prompt(f"Do you accept or reject the offer of ${offer} for you?"))
        choice = offer_json['choice']
        print(f"Selected: {choice} b/c {offer_json['explanation']}")
        # Aceept or reject
        self.driver.find_element(By.XPATH, f"//input[@value='{choice}']").click()
        # If reject create counter offer
        if choice == 'r':
            counter_offer_json = json.loads(self.agent.prompt("What is your counter offer? Decide between $1-3 to keep for yourself"))
            counter_offer = counter_offer_json['choice']
            print(f"Counter offer: {counter_offer} b/c {counter_offer_json['explanation']}")
            # Select counter offer
            select = Select(self.driver.find_element(By.NAME, "r_counter_proposal"))
            select.select_by_value(counter_offer)
        # Submit decision
        print("Submitting decision...")
        self.driver.find_element(By.XPATH, "//input[@value='Submit Response']").click()
    
        # Confirm decision
        self.wait.until(EC.url_contains("bg_r_confirm.php"))
        self.wait.until(EC.presence_of_element_located((By.XPATH, "//input[@value='Confirm Decision']")))
        print("Confirming decision...")
        self.driver.find_element(By.XPATH, "//input[@value='Confirm Decision']").click()
    
    def run(self):
        """
        Execute the full automation sequence.
        """
        try:
            self.login_step_1()
            self.wait.until(EC.url_contains("login2.php"))
            self.login_step_2()
            self.wait.until(EC.url_contains("inst_0.php"))

            self.save_instructions()
            self.wait.until(EC.url_contains("bg_inst1.php"))
            self.save_instructions()
            self.wait.until(EC.url_contains("bg_inst2.php"))
            # Check if first mover or second mover
            bold_elements = self.driver.find_elements(By.TAG_NAME, "b")
            first_mover = "Proposer" in " ".join([element.text for element in bold_elements])
            print(f"Agent is First mover: {first_mover}")
            self.save_instructions()
            self.wait.until(EC.url_contains("bg_inst6.php"))
            self.save_instructions()

            print("Reached decision page")
            # print("\n".join(self.instructions))
            if first_mover:
                self.play_game()
            else: # second mover (responder)
                self.respond_to_offer()

        finally:
            self.driver.quit()

def run_ultimatum_experiment(id: str, agents: int, password="pass"):
    """
    Run the ultimatum experiment.
    """
    # Run the automation
    def run_automation(first_name, last_name):
        automation = VeconLabAutomation(session_id=id, first_name=first_name, last_name=last_name, password=password)
        automation.run()

        # Create a log directory if it doesn't exist
        log_dir = os.path.join("logs", f"ult_log_{time.strftime('%Y-%m-%d-%H-%M')}")
        if not os.path.exists(log_dir):
            os.makedirs(log_dir)

        # Log the results
        with open(os.path.join(log_dir, f"{first_name}_{last_name}.txt"), "w") as log_file:
            # Write the interactions
            log_file.write("Interactions:\n")
            log_file.write(automation.previous_rounds.to_json(index=False))
            log_file.write("\n\n")

            # Write the agent's messages
            log_file.write("Agent Messages:\n")
            for message in automation.agent.messages:
                log_file.write(f"{message}\n")
            log_file.write("\n")
    
    # Start multiple threads for different players
    for i in range(agents):
        first_name = f"Agent"
        last_name = f"{i + 1}"
        thread = threading.Thread(target=run_automation, args=(first_name, last_name))
        thread.start()

# Example usage:
if __name__ == "__main__":
    automation = VeconLabAutomation(session_id="xaty30", first_name="Agent", last_name="One", password="1234")
    automation.run()



# Represents an agent, wrapper around a model

# Utility imports
import os
# LLM imports
from openai import OpenAI
# Formatting
from .prompts import format_prompt

class Agent:
    """
    An agent that interacts with a Large Language Model (LLM).

    This class provides a wrapper around a model provider (e.g., OpenAI),
    allowing the user to configure the model and securely pass/store API keys
    via multiple methods (arguments, environment variables, secrets file, etc).
    """

    def __init__(self, provider: str = "openai", model: str = "gpt-4o-mini", api_key: str = None, secrets_file: str = None):
        """
        Initialize the Agent with the specified provider, model, and API key.

        Parameters
        ----------
        provider : str
            The name of the LLM provider (default: "openai").
        model : str
            The model name or ID to be used (default: "gpt-4o-mini").
        api_key : str
            The API key to authenticate with the provider (default: None).
        secrets_file : str
            The path to a text file containing an "API_KEY" entry (default: None).
        
        Raises
        ------
        ValueError
            If no valid API key is provided or found in the environment/secrets file.

        Notes
        -----
        - Supported providers: "openai", "deepseek"
        - The API key can be passed as an argument, read from a secrets file, or an environment variable.

        Examples
        --------
        >>> agent = Agent(secrets_file="secrets.txt")
        """
        self.provider = provider.lower()
        self.model = model.lower()

        # Read the API key
        if api_key:
            self.api_key = api_key
        elif secrets_file: # read from secrets text file
            with open(secrets_file, "r") as f:
                # read the first line
                self.api_key = f.read().strip()
        else: # try reading from environment variable
            try:
                self.api_key = os.getenv("LLM_API_KEY")
            except:
                self.api_key = None
        if not self.api_key:
            raise ValueError("No valid API key provided or found in environment/secrets file.")
        
        # Create connection to the model provides
        if provider == "openai":
            self.client = OpenAI(api_key=self.api_key)
        elif provider == "deepseek":
            self.client = OpenAI(api_key=self.api_key, base_url="https://api.deepseek.com")
        else:
            raise ValueError(f"Unsupported provider: {provider}")
        
        # Initialize list of messages
        self.messages = []
    
    def add_context(self, context_text: str):
        """
        Adds context to the model. For OpenAI uses a developer prompt.

        Parameters
        ----------
        context_text : str
            The text to add to the context.
        """
        # OpenAI Completion API
        if self.provider == "openai":
            # Add the context to the list of messages
            formatted_msg = format_prompt(role="developer", prompt_text=context_text)
            self.messages.append(formatted_msg)

    def prompt(self, prompt_text: str) -> str:
        """
        Generate a response from the model given a prompt.

        Parameters
        ----------
        prompt_text : str
            The text prompt to generate a response for.

        Returns
        -------
        str
            The generated response text.

        Examples
        --------
        >>> response = agent.prompt("Once upon a time")
        """
        # OpenAI and DeepSeek completions
        if self.provider in ["openai", "deepseek"]:
            # Add the prompt to the list of messages
            formatted_msg = format_prompt(role="user", prompt_text=prompt_text)
            self.messages.append(formatted_msg)

            # Perform completion
            completion = self.client.chat.completions.create(
                model=self.model,
                messages=self.messages,
                max_tokens=100
            )

            # Extract the response
            result = completion.choices[0].message.content

            # Add the response to the list of messages
            formatted_msg = format_prompt(role="assistant", prompt_text=result)
            self.messages.append(formatted_msg)

            # Return the response
            return result
        else:
            raise ValueError(f"Unsupported provider: {self.provider}")
